package es.ehu.elecciones;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeMap;

import javax.swing.JFrame;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.data.general.DefaultPieDataset;

/**
 * Almacena y procesa los resultados de un proceso electoral.
 * 
 * El pais se supone dividido en circunscripciones y las circunscripciones
 * en colegios.
 * Partidos, colegios y circunscripciones se identifican por su codigo (string).
 * 
 * Un ComputadorVotos es capaz de producir los siguientes resultados:
 *    - el total de votos y porcentaje de cada partido
 *    (en el conjunto de circunscripciones)
 *    - el total de votos y porcentaje de cada partido en
 *    una circunscripcion dada
 * 
 * Un ComputadorVotos proporciona metodos para:
 *    - cargar datos de un proceso electoral
 *    - producir los resultados referidos
 *      ATENCION: NO pueden producirse resultados si antes no se han cargado
 *      los datos correspondientes.
 *      
 * Los datos de un proceso electoral se leen de un archivo de texto.
 * Cada linea del archivo representa el numero de votos de un partido
 * en un colegio electoral y debe contener:
 *     - el codigo de la circunscripcion (string)
 *     - el codigo del colegio (string)
 *     - el codigo del partido votado
 *     - el numero de votos obtenidos (posiblemente cero)
 */
public class ComputadorVotos
{
	private static final long serialVersionUID = 1L;

	public static final String INPUT_PATH = "resources/ResultadosElectorales-00.txt";
	
	public static void main(String[] args) {
		ComputadorVotos obj = new ComputadorVotos();
		obj.test();
	}
	
	private void test(){
		File file = null;
		Scanner scanner = null;
		file = new File(INPUT_PATH);
		try {
			scanner = new Scanner(file);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		cargarDatos(scanner);
		printInfoVotosPorPartido();
		printInfoVotosPorPartidoYCircunscripcion();
		
		verResultadosPorPartido();
		verResultadosCircunscripcion();
	}
	
	private static final String CABECERA_RESULTADOS_POR_PARTIDO =
		"*** Computo general por partido ***";

	
	/*
	 * Almacenamiento de la informacion de un proceso electoral
	 */
	
	/* Total de votos */
	private int totalVotos;

	/**
	 * Votos por cada partido en cada circunscripcion.
	 */
	private TreeMap<String, TreeMap<String, Integer>> votosPorPartidoYCircunscripcion = 
		new TreeMap<String, TreeMap<String, Integer>>();

	/**
	 * Votos por cada partido en el conjunto de circunscripciones.
	 */
	private TreeMap<String, Integer> votosPorPartido = new TreeMap<String, Integer>();

	/*
	 * un ChartPanel sirve para dibujar representaciones graficas de datos
	 */

	
	/**
	 * Lectura de los datos contenidos en el archivo correspondiente
	 * al Scanner 'input'.
	 */
	public void cargarDatos(Scanner input)
	{
		totalVotos = 0;
		votosPorPartidoYCircunscripcion.clear(); //vaciar de informacion previa
		votosPorPartido.clear();
		if (input != null){

			/*
			 *  bucle para lectura y procesamiento de datos
			 * 
			 *  el archivo de entrada debe ajustarse al formato indicado en el enunciado
			 */
			while (input.hasNext())
			{
				String cdgCircunscripcion = input.next();
				input.next();             // Saltar cdg de colegio
				String cdgPartido = input.next();
				int votos = input.nextInt();

				totalVotos = totalVotos + votos;
				contabilizarVotosPartidoEnCircunscripcion(cdgCircunscripcion,
				   		                                  cdgPartido,
						                                  votos);
				/*
				 * Los votos totales de cada partido estan almacenados en la tabla:
				 *           votosPorPartido
				 * Hay que cambiar el numero de votos registrados sumando 'votos'.
				 */
				sumarVotosAPartidoEnTabla(cdgPartido, votos, votosPorPartido);            
			}
		}
	}

	/**
	 * Devuelve una nueva tabla con los porcentajes de cada partido,
	 * en el total de las circunscripciones.
	 */
	public TreeMap<String, Double> getPorcentajesPorPartido() 
	{
		return getTablaPorcentajesPorPartido(votosPorPartido);
	}
	
	/**
	 * Devuelve una nueva tabla con los porcentajes de cada partido,
	 * en una circunscripcion dada.
	 */
	public TreeMap<String, Double> getPorcentajesPorPartidoEn(String circunscripcion) 
	{
		/*
		 * Resultados de cada partido en esa circunscripcion
		 */
		TreeMap<String, Integer> votosEnCircunscripcion =
			votosPorPartidoYCircunscripcion.get(circunscripcion);
		if (votosEnCircunscripcion == null) return null;
		else return getTablaPorcentajesPorPartido(votosEnCircunscripcion);
	}
	
	/**
	 * Se escriben en la salida standard los resultados obtenidos 
	 * por cada partido en el conjunto de cirunscripciones.
	 */
	public void printInfoVotosPorPartido() 
	{
		printInfoVotosPorPartido(System.out);
	}
	
	/**
	 * Se escriben en el archivo correspondiente al PrintStream 'out'
	 * los resultados obtenidos por cada partido en el conjunto de cirunscripciones.
	 */
	public void printInfoVotosPorPartido(PrintStream out) 
	{
		out.println(CABECERA_RESULTADOS_POR_PARTIDO);		
		printInfoVotosPorPartidoEnTabla(out, votosPorPartido);
	}

	/**
	 * Se escriben en la salida standard los resultados obtenidos 
	 * por cada partido en cada cirunscripcion.
	 */
	public void printInfoVotosPorPartidoYCircunscripcion()
	{
		printInfoVotosPorPartidoYCircunscripcion(System.out);
	}
	
	/**
	 * Se escriben en el archivo correspondiente al PrintStream 'out'
	 * los resultados obtenidos por cada partido en cada cirunscripcion.
	 */
	public void printInfoVotosPorPartidoYCircunscripcion(PrintStream out)
	{
		out.println("*** Computo por circunscripcion y partido ***");
		/*
		 * El conjunto de nombres de circunscripcion que estan en la tabla es...
		 */
		Set<String> conjuntoDeCircunscripciones = 
			votosPorPartidoYCircunscripcion.keySet();

		/* Para cada circunscripcion "c" de ese conjunto.. */
		for (String c: conjuntoDeCircunscripciones)
		{
			out.println("*** Resultados Parciales {" + c + "} ***");
			
			/*
			 * Resultados de cada partido en esa circunscripcion
			 */
			TreeMap<String, Integer> votosEnCircunscripcion =
				votosPorPartidoYCircunscripcion.get(c);
			
			printInfoVotosPorPartidoEnTabla(out, votosEnCircunscripcion);
		}
	}
	
	/**
	 * Se escriben en el archivo correspondiente al PrintStream 'out'
	 * los resultados registrados para cada partido en 'map'. 
	 */
	private void printInfoVotosPorPartidoEnTabla(PrintStream out,
			                                     TreeMap<String, Integer> map) 
	{
		/*
		 * El conjunto de nombres de partidos que estan en la tabla es...
		 */
		Set<String> conjuntoDePartidos = map.keySet();

		/* Para cada partido "p" del conjunto de partidos.. */
		for (String p: conjuntoDePartidos)
		{
			int n = map.get(p);
			out.println(p + " [" + n + "]");
		}		
	}

	/**
	 * Devuelve una nueva tabla con los porcentajes de cada partido,
	 * conforme a los datos registrados en 'map'
	 */
	private TreeMap<String, Double> getTablaPorcentajesPorPartido(
			TreeMap<String, Integer> map) 
	{
		TreeMap<String, Double> rsl = new TreeMap<String, Double>();
		
		/*
		 * El conjunto de nombres de partidos que estan en la tabla es...
		 */
		Set<String> conjuntoDePartidos = map.keySet();

		/* Para cada partido "p" del conjunto de partidos.. */
		for (String p: conjuntoDePartidos)
		{
			/* el numero de votos del partido 'p' es...*/
			int v = map.get(p);
			double porcentaje = 100 * v/((double)totalVotos);
			/* 
			 * el total de votos se convierte a double para evitar
			 * el redondeo de la division entera
			 */
			rsl.put(p, porcentaje);
		}
		return rsl;
	}


	/**
	 * Suma un numero de votos al 'partido' indicado en la 'circunscripcion'.
	 */
	private void contabilizarVotosPartidoEnCircunscripcion(
			String circunscripcion, 
			String partido, 
			int votos) 
	{
		/*
		 * Los votos de cada circunscripcion estan almacenados en la tabla:
		 *           votosPorPartidoYCircunscripcion
		 * Por cada circunscripcion, esa tabla contiene a su vez una tabla
		 * por cada partido, con los votos de cada uno.
		 * Hay que cambiar ese numero de votos, sumandole 'votos'.
		 */
		
		/*
		 * La tabla con los votos de cada partido en la 'circunscripcion' es...
		 */
		TreeMap<String, Integer> votosEnCircunscripcion =
			votosPorPartidoYCircunscripcion.get(circunscripcion);
		
		/*
		 * OBSERVACION:
		 * Puede ocurrir que 'circunscripcion' no este todavia registrado en la tabla
		 *       votosPorPartidoYCircunscripcion
		 * eso ocurre cuando se va a apuntar por primera vez un numero de votos
		 * a algun partido en esa circunscripcion.
		 */
		if (votosEnCircunscripcion == null){
			/*
			 * Circunscripcion sin registrar
			 * 
			 * Crear nueva tabla para registrar votos en esa circunscripcion
			 */
			votosEnCircunscripcion = new TreeMap<String, Integer>();

			/*
			 * Registrar en tabla de votos por partido y circunscripcion
			 */
			votosPorPartidoYCircunscripcion.put(circunscripcion, 
					                            votosEnCircunscripcion);
		}
		sumarVotosAPartidoEnTabla(partido, votos, votosEnCircunscripcion);
	}

	/**
	 * Suma un numero de votos al 'partido' indicado en una tabla con votos
	 * por partido.
	 */
	private void sumarVotosAPartidoEnTabla(String partido, int votos,
			                               TreeMap<String, Integer> tablaVotosPorPartido)
	{
		/*
		 * OBSERVACION:
		 * Puede ocurrir que 'partido' no este todavia registrado en la tabla;
		 * eso ocurre cuando se va a apuntar por primera vez un numero de votos
		 * a ese partido en esa tabla.
		 */
		int votosDelPartido = -1; //de entrada, es -1
		
		/*
		 * Ahora se examinara cual era el numero de votos hasta ahora
		 */
		if (tablaVotosPorPartido.get(partido) == null)
		{
			/*
			 * Partido sin registrar
			 * 
		     * VER NOTAS:En definitiva las tablas almacenan OBJETOS
		     * Si se trata de numeros, son objetos de clase Integer
			 */
			votosDelPartido = 0;
		} else {
			/*
			 * Ese partido ya tenia votos: recuperar ese numero
			 */
			votosDelPartido = votosPorPartido.get(partido);
		}
		votosDelPartido = votosDelPartido + votos;
		
		/*
		 * Actualizar estado de la tabla con los votos de cada partido
		 */
		tablaVotosPorPartido.put(partido, votosDelPartido);
	}


	/**
	 * Visualizacion grafica de los resultados obtenidos por cada partido en CADA 
	 * circunscripcion.
	 */
	public void verResultadosCircunscripcion()
	{
		/*
		 * Para cada par (clave, valor) del map...
		 */
        for (Map.Entry<String, TreeMap<String, Integer>> e: votosPorPartidoYCircunscripcion.entrySet()){
        	String key = e.getKey();
            
        	verResultadosCircunscripcion(key);
        }               
	}
	
	/**
	 * Visualizacion grafica de los resultados obtenidos por cada partido en UNA 
	 * circunscripcion.
	 */
	public void verResultadosCircunscripcion(String cirunscripcion)
	{
		/*
		 * La tabla con los porcentajes de cada partido en esa circunscripcion es...
		 */
		TreeMap<String, Double> map = getPorcentajesPorPartidoEn(cirunscripcion);
		if (map != null){
			/*
			 * Construir una representacion grafica de esos datos
			 */
			String titulo = "Porcentajes por partido(" + cirunscripcion + ")";
			show(titulo, map); 
		}
	}

    
	/**
	 * Visualizacion grafica de los resultados obtenidos por cada partido.
	 */
	public void verResultadosPorPartido()
	{
		/*
		 * La tabla con los porcentajes de cada partido es...
		 */
		TreeMap<String, Double> map = getPorcentajesPorPartido();
		/*
		 * Construir una representacion grafica de esos datos
		 */
		show("Generales", map);
	}

	/**
	 * Crear una ventana con una representacion grafica
	 * de los datos contenidos en un map.
	 * 
	 * Se usan objetos de clases definidas en "jfreechart".
	 * 
	 * @param titulo el titulo de la ventana.
	 * @param m el map referido.
	 */
	public void show(String titulo, TreeMap<String, Double> m)
    {
        DefaultPieDataset pieDataset = new DefaultPieDataset();
        
		/*
		 * Para cada par (clave, valor)...
		 */
        for (Map.Entry<String, Double> e: m.entrySet()){
        	String key = e.getKey();
            double v = e.getValue();
            
            pieDataset.setValue(key, v);
        }               
        
        JFreeChart chart = 
        	ChartFactory.createPieChart(titulo, pieDataset,
                                        true,          // legend
                                        true,          // tooltips
                                        false);        // URLs
    	ChartPanel panel = new ChartPanel(null);
    	panel.setChart(chart);
    	
    	JFrame frame = new JFrame();
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setTitle(titulo);
		frame.add("Center", panel);
		frame.pack();
		frame.setVisible(true);
   }

}